/*
 * WL24xx Wireless LAN PCMCIA Card Driver for Linux
 *
 * References:
 *
 *   1. WL24xx packet drivers (tooasm.asm)
 *   2. Access Point Firmware Interface Specification for IEEE 802.11 SUTRO
 *   3. IEEE 802.11
 *   4. Linux network driver (/usr/src/linux/drivers/net)
 *   5. ISA card driver - wl24.c
 *   6. Linux PCMCIA skeleton driver - skeleton.c
 *   7. Linux PCMCIA 3c589 network driver - 3c589_cs.c
 *
 * Program Flows:
 *
 *   Thread 1: init_module() -> wl24_attach() -> 
 *
 * Additional Information:
 *
 *   1. To prevent forward declaration, I put primitive functions ahead.
 *   2. All veriables are named with type, with form <type><name> e.g.:
 *      cTmp -> Type is 'c', and name is 'Tmp'
 *
 *      Type List: 
 *          c   - unsigned char (BYTE)
 *          u   - unsigned int  (WORD)
 *          b   - boolean int   (BOOL)
 *          ul  - unsigned long (DWORD)
 *          p   - pointer       (DWORD)
 *          psz - pointer to string ended with zero
 *          cpsz- const psz
 *          q   - queue
 *      Others are defined by Windows convension.
 *
 * Tested with WL2420 firmware 1.2, Linux 2.0.30, and pcmcia-cs-2.9.12
 *   1. Performance: about 165 Kbytes/sec in TCP/IP with Ad-Hoc mode.
 *      rsh 192.168.1.3 "dd if=/dev/zero bs=1k count=1000" > /dev/null
 *      (Specification 2M bits/sec. is about 250 Kbytes/sec., but we must deduct
 *       ETHER/IP/UDP/TCP header, and acknowledgement overhead)
 *
 */

/*======================================================================

    A wl24 PCMCIA client driver

    Written by Fox Chen, mhchen@golf.ccl.itri.org.tw
    
======================================================================*/

#include <pcmcia/config.h>
#include <pcmcia/k_compat.h>

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/ptrace.h>
#include <linux/malloc.h>
#include <linux/string.h>
#include <linux/timer.h>
#include <asm/io.h>
#include <asm/system.h>

#include <pcmcia/version.h>
#include <pcmcia/cs_types.h>
#include <pcmcia/cs.h>
#include <pcmcia/cistpl.h>
#include <pcmcia/ds.h>

#include "./wlapi.h"
#include "./wl24.h"
#include "./wlioctl.h"

/*
   All the PCMCIA modules use PCMCIA_DEBUG to control debugging.  If
   you do not define PCMCIA_DEBUG at all, all the debug code will be
   left out.  If you compile with PCMCIA_DEBUG=0, the debug code will
   be present but disabled -- but it can then be enabled for specific
   modules at load time with a 'pc_debug=#' option to insmod.
*/
#ifdef PCMCIA_DEBUG
    static int pc_debug = PCMCIA_DEBUG;
    MODULE_PARM(pc_debug, "i");
#   define DEBUG(n, format, args...) if (pc_debug>(n)) printk(format , ## args);
#else
#   define DEBUG(n, args...)
#endif

/*====================================================================*/

/* Parameters that can be set with 'insmod' */

/* Bit map of interrupts to choose from */
/* This means pick from 15, 14, 12, 11, 10, 9, 7, 5, 4, and 3 */
static u_long irq_mask = 0xdeb8;
static int irq_list[4] = { -1 };
static char *networkname = NULL;
static int networktype = Infrastructure;
static int Channel = 1;
//static int RTSThreshold = 2300;
static int LLCType = 1;     

MODULE_PARM(irq_mask, "i");
MODULE_PARM(irq_list, "1-4i");
MODULE_PARM(networkname, "s");
//MODULE_PARM_DESC(networkname, "ESSID (<string>) [ANY]");
MODULE_PARM(networktype, "b");
//MODULE_PARM_DESC(networktype, "NetworkType (0 - 1) [0]");
MODULE_PARM(Channel, "b");
//MODULE_PARM_DESC(Channel, "DefaultChannel (1 - 14) [1]");
MODULE_PARM(RTSThreshold, "b");
//MODULE_PARM_DESC(RTSThreshold, "RTSThreshold (50 - 2300) [2300]");
MODULE_PARM(LLCType, "b");
//MODULE_PARM_DESC(LLCType, "LLCType (1 - 2) [1]");

/*====================================================================*/

/*
   The event() function is this driver's Card Services event handler.
   It will be called by Card Services when an appropriate card status
   event is received.  The config() and release() entry points are
   used to configure or release a socket, in response to card insertion
   and ejection events.  They are invoked from the wl24 event
   handler.
*/

static void wl24_config(dev_link_t *link);
static void wl24_release(u_long arg);
static int wl24_event(event_t event, int priority, event_callback_args_t *args);

/*
   The attach() and detach() entry points are used to create and destroy
   "instances" of the driver, where each instance represents everything
   needed to manage one actual PCMCIA card.
*/

static dev_link_t *wl24_attach(void);
static void wl24_detach(dev_link_t *);

/*
   You'll also need to prototype all the functions that will actually
   be used to talk to your device.  See 'pcmem_cs' for a good example
   of a fully self-sufficient driver; the other drivers rely more or
   less on other parts of the kernel.
*/

void wl24_interrupt(int reg);

/*
   The dev_info variable is the "key" that is used to match up this
   device driver with appropriate cards, through the card configuration
   database.
*/

static dev_info_t dev_info = "wl24_cs";

/*
   A linked list of "instances" of the wl24 device.  Each actual
   PCMCIA card corresponds to one device instance, and is described
   by one dev_link_t structure (defined in ds.h).

   You may not want to use a linked list for this -- for example, the
   memory card driver uses an array of dev_link_t pointers, where minor
   device numbers are used to derive the corresponding array index.
*/

static dev_link_t *dev_list = NULL;

/*
   A dev_link_t structure has fields for most things that are needed
   to keep track of a socket, but there will usually be some device
   specific information that also needs to be kept track of.  The
   'priv' pointer in a dev_link_t structure can be used to point to
   a device-speific private data structure, like this.

   A driver needs to provide a dev_node_t structure for each device
   on a card.  In some cases, there is only one device per card (for
   example, ethernet cards, modems).  In other cases, there may be
   many actual or logical devices (SCSI adapters, memory cards with
   multiple partitions).  The dev_node_t structures need to be kept
   in a linked list starting at the 'dev' field of a dev_link_t
   structure.  We allocate them in the card's private data structure,
   because they generally can't be allocated dynamically.
   
typedef struct local_info_t {
    dev_node_t        node;
} local_info_t;
*/

/*====================================================================*/

static void cs_error(client_handle_t handle, int func, int ret)
{
    error_info_t err = { func, ret };
    CardServices(ReportError, handle, &err);
}

static int wl24_close(struct device *dev)
{
    WL_Adaptor *pAdaptor    = (WL_Adaptor *) dev->priv;
    dev_link_t *link;

    /* Check if the device is in dev_list */
    for (link = dev_list; link != NULL; link = link->next)
        if (link->priv == dev)
            break;
    if (link == NULL)
        return -ENODEV;

    link->open--;
    
    /* Stop WL_StartXmit() from now on */
    dev->tbusy = 1;
    dev->start = 0;
    pAdaptor->CardStart=FALSE;
    /* Ack Interrupt */
//    OutB(GCR_ECINT, pAdaptor->nBaseAddr + NIC_GCR);
    
    /* Mask interrupts from the SUTRO */
//    WL_BlockInterrupt(pAdaptor);

    if (link->state & DEV_STALE_CONFIG) {
        link->release.expires = RUN_AT(HZ/20);
        link->state |= DEV_RELEASE_PENDING;
        add_timer(&link->release);
    }
    
    MOD_DEC_USE_COUNT;
    
//   printk("%s : WL2431 closed\n", dev->name);
    return 0;
}

static int wl24_open(struct device *dev)
{
    //BYTE        cFilter, cCCR;
    WORD i;
    WL_Adaptor *pAdaptor    = (WL_Adaptor *) dev->priv;
    dev_link_t *link;

    /* Check if the device is in dev_list */
    for (link = dev_list; link != NULL; link = link->next)
        if (link->priv == dev)
            break;
    if (!DEV_OK(link))
        return -ENODEV;

    link->open++;
    MOD_INC_USE_COUNT;
    
    /* Initial WL24 firmware */
    printk("%s: Initialize WL24xx firmware...\n", dev->name);
    if (!WL_InitFirmware(pAdaptor)) {
        printk("%s: Cannot Initial Firmware!!\n", dev->name);
        return -ENODEV;
    }

    /* Initial device variables */
    dev->interrupt  = 0;
    dev->tbusy      = 0;
    dev->start      = 1;
    pAdaptor->CardStart=FALSE;
    pAdaptor->AdHocTimes=0; 
    /* Acknowledge Interrupt, for cleaning last state */
    OutB(GCR_ECINT, pAdaptor->nBaseAddr + NIC_GCR);
    
    /* Enable interrupt from card after all */
    WL_UnBlockInterrupt(pAdaptor);
    for (i=0;i<3000;i++){
    if (MgmtScan(pAdaptor,100)) break;
    } 
//    printk("%s: WL2431 opened\n", dev->name);
    return 0;
}

/*======================================================================

    We never need to do anything when a wl24 device is "initialized"
    by the net software, because we only register already-found cards.
    
======================================================================*/

static int wl24_init(struct device *dev)
{
    DEBUG(0, "wl24_init() called\n");
    return 0;
}

static int WL_Config(struct device *dev, struct ifmap *map)
{
    DEBUG(0, "WL_Config() called\n");
    return 0;
}

/*======================================================================

    wl24_attach() creates an "instance" of the driver, allocating
    local data structures for one device.  The device is registered
    with Card Services.

    The dev_link structure is initialized, but we don't actually
    configure the card at this point -- we wait until we receive a
    card insertion event.
    
======================================================================*/
static dev_link_t *wl24_attach(void)
{
    client_reg_t client_reg;
    dev_link_t *link;
    struct device *dev;
    int ret, i;
    
    DEBUG(0, "wl24_attach()\n");

    /* Initialize the dev_link_t structure */
    link = kmalloc(sizeof(struct dev_link_t), GFP_KERNEL);
    memset(link, 0, sizeof(struct dev_link_t));
    link->release.function = &wl24_release;
    link->release.data = (u_long)link;

    /* The io structure describes IO port mapping */
    link->io.NumPorts1 = 16;
    link->io.Attributes1 = IO_DATA_PATH_WIDTH_8;
    link->io.IOAddrLines = 5;

    /* Interrupt setup */
    link->irq.Attributes = IRQ_TYPE_EXCLUSIVE | IRQ_HANDLE_PRESENT;
    link->irq.IRQInfo1 = IRQ_INFO2_VALID|IRQ_LEVEL_ID;
    link->irq.IRQInfo2 = irq_mask;
    if (irq_list[0] == -1)
        link->irq.IRQInfo2 = irq_mask;
    else
        for (i = 0; i < 4; i++)
            link->irq.IRQInfo2 |= 1 << irq_list[i];
    link->irq.Handler = WL_Interrupt;
    
    /* General socket configuration */
    link->conf.Attributes = CONF_ENABLE_IRQ;
    link->conf.Vcc = 50;
    link->conf.IntType = INT_MEMORY_AND_IO;
    link->conf.ConfigIndex = 1;
    link->conf.Present = PRESENT_OPTION;

    /* Allocate space for private device-specific data */
    dev = kmalloc(sizeof(struct device), GFP_KERNEL);
    memset(dev, 0, sizeof(struct device));

    /* Store the WL_Adaptor sturcture into private memory */
    dev->priv = kmalloc(sizeof(WL_Adaptor), GFP_KERNEL);
    ASSERT(dev->priv != NULL);
        
    memset(dev->priv, 0, sizeof(WL_Adaptor));
    
    dev->hard_start_xmit    = &WL_StartXmit;
    dev->set_config         = &WL_Config;
    dev->get_stats          = &WL_GetStats;
    dev->set_multicast_list = &WL_SetMulticastList;
#ifdef __WL_IOCTL_H__
    dev->do_ioctl           = &WL_Ioctl;
#endif

    ether_setup(dev);
    dev->name = ((WL_Adaptor *)dev->priv)->node.dev_name;

    /* Set up Callback functions */
    dev->init               = &wl24_init;
    dev->open               = &wl24_open;
    dev->stop               = &wl24_close;

    dev->tbusy = 1;
    link->priv = link->irq.Instance = dev;
    
    /* Register with Card Services */
    link->next = dev_list;
    dev_list = link;
    client_reg.dev_info = &dev_info;
    client_reg.Attributes = INFO_IO_CLIENT | INFO_CARD_SHARE;
    client_reg.EventMask =
        CS_EVENT_CARD_INSERTION | CS_EVENT_CARD_REMOVAL |
        CS_EVENT_RESET_PHYSICAL | CS_EVENT_CARD_RESET |
        CS_EVENT_PM_SUSPEND | CS_EVENT_PM_RESUME;
    client_reg.event_handler = &wl24_event;
    client_reg.Version = 0x0210;
    client_reg.event_callback_args.client_data = link;
    ret = CardServices(RegisterClient, &link->handle, &client_reg);
    if (ret != 0) {
        cs_error(link->handle, RegisterClient, ret);
        wl24_detach(link);
        return NULL;
    }

    return link;
} /* wl24_attach */

/*======================================================================

    This deletes a driver "instance".  The device is de-registered
    with Card Services.  If it has been released, all local data
    structures are freed.  Otherwise, the structures will be freed
    when the device is released.

======================================================================*/

static void wl24_detach(dev_link_t *link)
{
    dev_link_t **linkp;

    DEBUG(0, "wl24_detach(0x%p)\n", link);
    
    /* Locate device structure */
    for (linkp = &dev_list; *linkp; linkp = &(*linkp)->next)
        if (*linkp == link) break;
    if (*linkp == NULL)
        return;

    /*
       If the device is currently configured and active, we won't
       actually delete it yet.  Instead, it is marked so that when
       the release() function is called, that will trigger a proper
       detach().
    */
    if (link->state & DEV_CONFIG) {
#ifdef PCMCIA_DEBUG
        printk(KERN_DEBUG "wl24_cs: detach postponed, '%s' "
               "still locked\n", link->dev->dev_name);
#endif
        link->state |= DEV_STALE_LINK;
        return;
    }

    /* Break the link with Card Services */
    if (link->handle)
        CardServices(DeregisterClient, link->handle);
    
    /* Unlink device structure, free pieces */
    *linkp = link->next;
    if (link->priv) {
        struct device *dev = link->priv;
        if (dev->priv)
            kfree_s(dev->priv, sizeof(WL_Adaptor));
        kfree_s(link->priv, sizeof(struct device));
    }
    kfree_s(link, sizeof(struct dev_link_t));
    
} /* wl24_detach */

/*======================================================================

    wl24_config() is scheduled to run after a CARD_INSERTION event
    is received, to configure the PCMCIA socket, and to make the
    ethernet device available to the system.
    
======================================================================*/

#define CS_CHECK(fn, args...) \
while ((last_ret=CardServices(last_fn=(fn),args))!=0) goto cs_failed

static void wl24_config(dev_link_t *link)
{
    client_handle_t handle;
    tuple_t tuple;
    cisparse_t parse;
    struct device *dev;
    int i, j, last_fn, last_ret;
    u_char buf[64];
    WL_Adaptor *pAdaptor;
    
    handle = link->handle;
    dev = link->priv;

    DEBUG(0, "wl24_config(0x%p)\n", link);

    /*
       This reads the card's CONFIG tuple to find its configuration
       registers.
    */
    tuple.Attributes = 0;                           /* diff */
    tuple.DesiredTuple = CISTPL_CONFIG;
    CS_CHECK(GetFirstTuple, handle, &tuple);
    tuple.TupleData = buf;
    tuple.TupleDataMax = 64;
    tuple.TupleOffset = 0;
    CS_CHECK(GetTupleData, handle, &tuple);
    CS_CHECK(ParseTuple, handle, &tuple, &parse);
    link->conf.ConfigBase = parse.config.base;
    link->conf.Present = parse.config.rmask[0];     /*diff */
    
    /* Configure card */
    link->state |= DEV_CONFIG;

    /*
       Try allocating IO ports.  This tries a few fixed addresses.
       If you want, you can also read the card's config table to
       pick addresses -- see the serial driver for an example.
    */
    for (j = 0x280; j < 0x400; j += 0x20) {
        /* The '^0x300' is so that we probe 0x300-0x3ff first, then
           0x200-0x2ff, and so on, because this seems safer */
        link->io.BasePort1 = j;
        link->io.BasePort2 = link->io.BasePort1 + 0x20;
        i = CardServices(RequestIO, link->handle, &link->io);
        if (i == CS_SUCCESS) break;
    }
    if (i != CS_SUCCESS) {
        cs_error(link->handle, RequestIO, i);
        goto failed;
    }
    
    /*
       Now allocate an interrupt line.  Note that this does not
       actually assign a handler to the interrupt.
    */
    CS_CHECK(RequestIRQ, link->handle, &link->irq);
        
    /*
       This actually configures the PCMCIA socket -- setting up
       the I/O windows and the interrupt mapping.
    */
    CS_CHECK(RequestConfiguration, link->handle, &link->conf);

    dev->irq = link->irq.AssignedIRQ;
    dev->base_addr = link->io.BasePort1;
    dev->tbusy = 0;
    if (register_netdev(dev) != 0) {
        printk(KERN_NOTICE "wl24_cs: register_netdev() failed\n");
        goto failed;
    }
    
    /* At this point, the dev_node_t structure(s) should be
       initialized and arranged in a linked list at link->dev. */
    link->dev = &((WL_Adaptor *)dev->priv)->node;
    link->state &= ~DEV_CONFIG_PENDING;

    pAdaptor = (WL_Adaptor *) dev->priv;
    pAdaptor->nBaseAddr = dev->base_addr;
    if (!WL_GetFlashMacAddress(pAdaptor)) {
        printk("%s: Cannot read MAC address in flash ROM ??\n", dev->name);
        goto failed;
    }

    /* print probe information */
    printk("%s: WL24xx at 0x%3.3x, IRQ %d, MAC address in flash ROM:", 
        dev->name, pAdaptor->nBaseAddr, (int) dev->irq);
    for (i = 0; i < 6; i++) {
        dev->dev_addr[i] = ((BYTE *) &pAdaptor->MacAddress)[i];
        printk("%c%02x", i ? ':' : ' ', dev->dev_addr[i]);
    }
    printk("\n");
// initila adaptor parameter // add by jss
    pAdaptor->NetworkType = networktype;
    pAdaptor->LLCType = LLCType;
    pAdaptor->DefaultChannel = Channel;
    pAdaptor->BSSCount = 0;
    pAdaptor->JoinStaBSS = 0;
    pAdaptor->AdHocTimes = 0;
    pAdaptor->DriverState = 0;
    pAdaptor->CardStart = FALSE;
    pAdaptor->ESSID[0] = 0;
    if ((!networkname) || (!networkname[0]))
    {
      pAdaptor->ESSID[1] = 3;
      pAdaptor->ESSID[2] = 'A';
      pAdaptor->ESSID[3] = 'N';
      pAdaptor->ESSID[4] = 'Y';
    }
    else
    {
      pAdaptor->ESSID[1] = strlen(networkname);
      strcpy(pAdaptor->ESSID + 2, networkname);
    }
    switch (pAdaptor->FrequencyDomain){
        case 0x10:
        if ((pAdaptor->DefaultChannel < 1) || (pAdaptor->DefaultChannel>11))
                pAdaptor->DefaultChannel = 1;
            break;
        case 0x20:
        if ((pAdaptor->DefaultChannel < 1) || (pAdaptor->DefaultChannel>11))
                pAdaptor->DefaultChannel = 1;
            break;
        case 0x30:
        if ((pAdaptor->DefaultChannel < 1) || (pAdaptor->DefaultChannel>13))
                pAdaptor->DefaultChannel = 1;
            break;
        case 0x31:
        if ((pAdaptor->DefaultChannel < 10) || (pAdaptor->DefaultChannel>11))
                pAdaptor->DefaultChannel = 10;
            break;
        case 0x32:
        if ((pAdaptor->DefaultChannel < 10) || (pAdaptor->DefaultChannel>13))
                pAdaptor->DefaultChannel = 10;
            break;
        case 0x40:
            pAdaptor->DefaultChannel = 14;
            break;
        default:
            pAdaptor->DefaultChannel = 1;
            break;
    }

    return;

cs_failed:
    cs_error(link->handle, last_fn, last_ret);
failed:
    wl24_release((u_long)link);

} /* wl24_config */

/*======================================================================

    After a card is removed, wl24_release() will unregister the net
    device, and release the PCMCIA configuration.  If the device is
    still open, this will be postponed until it is closed.
    
======================================================================*/

static void wl24_release(u_long arg)
{
    dev_link_t *link = (dev_link_t *)arg;
    struct device *dev = link->priv;

    DEBUG(0, "wl24_release(0x%p)\n", link);

    /*
       If the device is currently in use, we won't release until it
       is actually closed.
    */
    if (link->open) {
        DEBUG(1, "wl24_cs: release postponed, '%s' still open\n",
              link->dev->dev_name);
        link->state |= DEV_STALE_CONFIG;
        return;
    }

    /* Unlink the device chain */
    if (link->dev != NULL) {
        unregister_netdev(dev);
        link->dev = NULL;
    }
    
    /* Don't bother checking to see if these succeed or not */
    CardServices(ReleaseConfiguration, link->handle);
    CardServices(ReleaseIO, link->handle, &link->io);
    CardServices(ReleaseIRQ, link->handle, &link->irq);
    link->state &= ~DEV_CONFIG;
    
    if (link->state & DEV_STALE_LINK)
        wl24_detach(link);
    
} /* wl24_release */

/*======================================================================

    The card status event handler.  Mostly, this schedules other
    stuff to run after an event is received.  A CARD_REMOVAL event
    also sets some flags to discourage the net drivers from trying
    to talk to the card any more.

    When a CARD_REMOVAL event is received, we immediately set a flag
    to block future accesses to this device.  All the functions that
    actually access the device should check this flag to make sure
    the card is still present.
    
======================================================================*/

static int wl24_event(event_t event, int priority, event_callback_args_t *args)
{
    dev_link_t *link = args->client_data;
    struct device *dev = link->priv;

    DEBUG(1, "wl24_event(0x%06x)\n", event);
    
    switch (event) {
    case CS_EVENT_CARD_REMOVAL:
        link->state &= ~DEV_PRESENT;
        if (link->state & DEV_CONFIG) {
            dev->tbusy = 1;
            dev->start = 0;
            link->release.expires = RUN_AT(HZ/20);
            add_timer(&link->release);
        }
        break;
    case CS_EVENT_CARD_INSERTION:
        link->state |= DEV_PRESENT | DEV_CONFIG_PENDING;
        wl24_config(link);
        break;
    case CS_EVENT_PM_SUSPEND:
        link->state |= DEV_SUSPEND;
        /* Fall through... */
    case CS_EVENT_RESET_PHYSICAL:
        if (link->state & DEV_CONFIG) {
            if (link->open) {
                dev->tbusy = 1; dev->start = 0;
            }
            CardServices(ReleaseConfiguration, link->handle);
        }
        break;
    case CS_EVENT_PM_RESUME:
        link->state &= ~DEV_SUSPEND;
        /* Fall through... */
    case CS_EVENT_CARD_RESET:
        if (link->state & DEV_CONFIG) {
            CardServices(RequestConfiguration, link->handle, &link->conf);
            if (link->open) {
                WL_Reset(dev);
            }
        }
        break;
    }
    return 0;
} /* wl24_event */

/*====================================================================*/
int init_module(void)
{
    servinfo_t serv;
    CardServices(GetCardServicesInfo, &serv);
    if (serv.Revision != CS_RELEASE_CODE) {
        printk(KERN_NOTICE "wl24_cs: Card Services release does not match!\n"
               "Compiled with 0x%x, but current is 0x%lx\n", 
               CS_RELEASE_CODE, (long)serv.Revision);
        /* return -1; */
    }
    register_pcmcia_driver(&dev_info, &wl24_attach, &wl24_detach);
    return 0;
}

void cleanup_module(void)
{
    DEBUG(0, "wl24_cs: unloading\n");
    unregister_pcmcia_driver(&dev_info);
    while (dev_list != NULL) {
        if (dev_list->state & DEV_CONFIG)
            wl24_release((u_long)dev_list);
        wl24_detach(dev_list);
    }
}

