/*
 * Refer AMD Am29F010 CMOS Sector Erase Flash Memory Data Sheet.
 */

#include "./wlapi.h"
#include "./wlioctl.h"

#include <asm/segment.h>
#if LINUX_VERSION_CODE >= 0x20200
#include <asm/uaccess.h>
#endif

/* Hack some linux functions */
#ifdef __FreeBSD__
   extern const char g_cpszVersion[];
   struct zwl_softc;
   extern int zwl_reset(struct zwl_softc *sc);
#  define WL_Reset(sc)      zwl_reset(sc)
#  define memcpy_tofs(pDest, pSrc, ulLen)       copyout(pSrc, pDest, ulLen)
#  define memcpy_fromfs(pDest, pSrc, ulLen)     copyin(pSrc, pDest, ulLen)
#  define suser()   (suser(curproc->p_ucred, &curproc->p_acflag) == 0)
#  include <sys/kernel.h>       /* for ticks */
#  define WL_DELTA_TIME(x)      ((x) = ticks - (x))

#  define VERIFY_READ   0
#  define VERIFY_WRITE  1

static int verify_area(int nOperation, const void *pData, WORD ulLen)
{
    /* always success */
    return 0;
}

#else /* For Linux */

#  include "wl24.h"
/* extern const char g_cpszVersion[]; */
/* extern int WL_Reset(struct device *dev); */

#  define WL_DELTA_TIME(x)      ((x) = jiffies - (x))

/*
 * Perform IOCTL call functions here. Some are privileged operations and the
 * effective uid is checked in those cases.
 *
 * This part is optional. Needed only if you want to run wlu (unix version).
 *
 * CAUTION: To prevent interrupted by WL_Interrupt() and timer-based 
 * WL_StartXmit() from other interrupts, this should be run single-threaded.
 * This function is expected to be a rare operation, and it's 
 * simpler to just use cli() to disable ALL interrupts.
 */
int WL_Ioctl(struct device *dev, struct ifreq *rq, int cmd)
{
    WL_Adaptor      *pAdaptor = (WL_Adaptor *) dev->priv;
    WL_IOCTL_BLK    *pBlk = (WL_IOCTL_BLK *) &rq->ifr_data;
    unsigned long   ulFlags;
    int             nReturn;

    /* Keep flags first, these are inline functions */
    save_flags(ulFlags);
    cli();

    nReturn = WL_IoctlWla(pAdaptor, pBlk, dev);
    
    /* Restore CLI and others */
    restore_flags(ulFlags);

    return nReturn;
}

#endif /* Linux End */

/*
 * Common ioctl part for both Linux and FreeBSD
 * Return 0 if success.
 */
int WL_IoctlWla(WL_Adaptor *pAdaptor, WL_IOCTL_BLK *pBlk, void *pDevice)
{
    int  nReturn;
    WL_IOCTL_PARA lParam;
    unsigned char buff[1028];
    switch(pBlk->uCmd) {
      case WL_IOCTL_SET_RESET:      /* Reset driver (needed after set) */
        if (!suser()) {
            nReturn = -EPERM;
            break;
        }
        
        nReturn = WL_Reset(pDevice);
        break;

      case WL_IOCTL_WRITE_FLASH:      /* Write firmware into Flash ROM */
        if (!suser()) {
            nReturn = -EPERM;
            break;
        }

#if LINUX_VERSION_CODE >= 0x20200
        copy_from_user( buff, pBlk->pData, pBlk->uLen);
#else
        memcpy_fromfs( buff, pBlk->pData, pBlk->uLen);
#endif
        nReturn = WL_WriteFlash(pAdaptor,buff, pBlk->uLen) ? 0 : -EIO;
        break;

      case WL_IOCTL_GET_PARAMETER: {  /* Get parameter */
        lParam.DefaultChannel = pAdaptor->DefaultChannel;
        lParam.Channel = pAdaptor->Channel;
        lParam.NetworkType = pAdaptor->NetworkType;
        lParam.LLCType = pAdaptor->LLCType;
        lParam.Version[0] = pAdaptor->Version[0];
        lParam.Version[1] = pAdaptor->Version[1];
        lParam.FrequencyDomain = pAdaptor->FrequencyDomain;
        memcpy((char *) &(lParam.KeepESSID[0]),(char *) &(pAdaptor->KeepESSID[0]), 34);
        memcpy((char *) &(lParam.ESSID[0]), (char *) &(pAdaptor->ESSID[0]), 34);
        pBlk->uLen = sizeof(lParam);
#if LINUX_VERSION_CODE >= 0x20200
        copy_to_user(pBlk->pData, &lParam, pBlk->uLen);
#else
        memcpy_tofs(pBlk->pData, &lParam, pBlk->uLen);
#endif
        return 0;
      }

      case WL_IOCTL_SET_PARAMETER: {  /* Set parameter */
        if (!suser()) {
            nReturn = -EPERM;
            break;
        }
        nReturn = verify_area(VERIFY_READ, (void *)pBlk->pData, sizeof(lParam));

        if (nReturn != 0)
            break;

#if LINUX_VERSION_CODE >= 0x20200
        copy_from_user( &lParam, pBlk->pData, sizeof(lParam));
#else
        memcpy_fromfs( &lParam, pBlk->pData, sizeof(lParam));
#endif
        pAdaptor->DefaultChannel=lParam.DefaultChannel;
        pAdaptor->NetworkType=lParam.NetworkType;
        pAdaptor->LLCType=lParam.LLCType;
        memcpy((char *) &(pAdaptor->ESSID[0]), (char *) &(lParam.ESSID[0]), 34); 
        nReturn = WL_Reset(pDevice);
        break;
      }
      default:
        nReturn = -EOPNOTSUPP;
    }
    
    return nReturn;
}

